@@ -5,6 +5,8 @@ from PIL import Image as PILImage
 
 import numpy as np
 
+import orjson
+
 
 def image_bytes_to_array(im_bytes):
     """Turn raw image bytes into a NumPy array."""
@@ -67,7 +69,9 @@ def populate_args(arg, args, buffers):
         args.append(arg)
 
 
-def to_camel_case(snake_str):
-    """Snake case to Camel case translator."""
-    components = snake_str.split('_')
-    return components[0] + ''.join(x.title() for x in components[1:])
+def commands_to_buffer(commands):
+    # Turn the commands list into a binary buffer
+    return array_to_binary(np.frombuffer(
+        bytes(orjson.dumps(commands, option=orjson.OPT_SERIALIZE_NUMPY)),
+        dtype=np.uint8)
+    )
