@@ -1,8 +1,11 @@
-function getTypedArray(dataview: any, metadata: any) {
+export function getTypedArray(dataview: any, metadata: any) {
   switch (metadata.dtype) {
     case 'int8':
       return new Int8Array(dataview.buffer);
       break;
+    case 'uint8':
+        return new Uint8Array(dataview.buffer);
+        break;
     case 'int16':
       return new Int16Array(dataview.buffer);
       break;
