@@ -14,7 +14,7 @@ import {
 } from './version';
 
 import {
-  getArg, toBytes, fromBytes
+  getArg, toBytes, fromBytes, getTypedArray
 } from './utils';
 
 
@@ -38,6 +38,41 @@ function deserializeImageData(dataview: DataView | null) {
   return new Uint8ClampedArray(dataview.buffer);
 }
 
+async function createImageFromWidget(image: DOMWidgetModel): Promise<HTMLImageElement> {
+  // Create the image manually instead of creating an ImageView
+  let url: string;
+  const format = image.get('format');
+  const value = image.get('value');
+  if (format !== 'url') {
+      const blob = new Blob([value], {type: `image/${format}`});
+      url = URL.createObjectURL(blob);
+  } else {
+      url = (new TextDecoder('utf-8')).decode(value.buffer);
+  }
+
+  const img = new Image();
+  return new Promise((resolve) => {
+    img.onload = () => {
+      resolve(img);
+    };
+    img.src = url;
+  });
+}
+
+
+const COMMANDS = [
+  'fillRect', 'strokeRect', 'fillRects', 'strokeRects', 'clearRect', 'fillArc',
+  'fillCircle', 'strokeArc', 'strokeCircle', 'fillArcs', 'strokeArcs',
+  'fillCircles', 'strokeCircles', 'strokeLine', 'beginPath', 'closePath',
+  'stroke', 'fillPath', 'fill', 'moveTo', 'lineTo',
+  'rect', 'arc', 'ellipse', 'arcTo', 'quadraticCurveTo',
+  'bezierCurveTo', 'fillText', 'strokeText', 'setLineDash', 'drawImage',
+  'putImageData', 'clip', 'save', 'restore', 'translate',
+  'rotate', 'scale', 'transform', 'setTransform', 'resetTransform',
+  'set', 'clear', 'sleep', 'fillPolygon', 'strokePolygon',
+  'strokeLines',
+];
+
 
 export
 class Path2DModel extends WidgetModel {
@@ -46,9 +81,6 @@ class Path2DModel extends WidgetModel {
       _model_name: Path2DModel.model_name,
       _model_module: Path2DModel.model_module,
       _model_module_version: Path2DModel.model_module_version,
-      _view_name: Path2DModel.view_name,
-      _view_module: Path2DModel.view_module,
-      _view_module_version: Path2DModel.view_module_version,
       value: '',
     };
   }
@@ -64,9 +96,135 @@ class Path2DModel extends WidgetModel {
   static model_name = 'Path2DModel';
   static model_module = MODULE_NAME;
   static model_module_version = MODULE_VERSION;
-  static view_name = 'Path2DView';
-  static view_module = MODULE_NAME;
-  static view_module_version = MODULE_VERSION;
+}
+
+
+export
+class PatternModel extends WidgetModel {
+  defaults() {
+    return {...super.defaults(),
+      _model_name: PatternModel.model_name,
+      _model_module: PatternModel.model_module,
+      _model_module_version: PatternModel.model_module_version,
+      image: '',
+      repetition: 'repeat',
+    };
+  }
+
+  async initialize(attributes: any, options: any) {
+    super.initialize(attributes, options);
+
+    const image = this.get('image');
+    let patternSource: HTMLCanvasElement | HTMLImageElement | undefined = undefined;
+
+    if (image instanceof CanvasModel || image instanceof MultiCanvasModel) {
+      patternSource = image.canvas;
+    }
+
+    if (image.get('_model_name') == 'ImageModel') {
+      const img = await createImageFromWidget(image);
+      patternSource = img;
+    }
+
+    if (patternSource == undefined) {
+      throw "Could not understand the souce for the pattern";
+    }
+
+    const pattern = PatternModel.ctx.createPattern(patternSource, this.get('repetition'));
+
+    if (pattern == null) {
+      throw "Could not initialize pattern object";
+    }
+
+    this.value = pattern;
+  }
+
+  static serializers: ISerializers = {
+    ...WidgetModel.serializers,
+    image: { deserialize: (unpack_models as any) },
+  }
+
+  value: CanvasPattern;
+
+  static model_name = 'PatternModel';
+  static model_module = MODULE_NAME;
+  static model_module_version = MODULE_VERSION;
+
+  // Global context for creating the gradients
+  static ctx: CanvasRenderingContext2D = getContext(document.createElement('canvas'));
+}
+
+
+class GradientModel extends WidgetModel {
+  defaults() {
+    return {...super.defaults(),
+      _model_module: GradientModel.model_module,
+      _model_module_version: GradientModel.model_module_version,
+      x0: 0.,
+      y0: 0.,
+      x1: 0.,
+      y1: 0.,
+      color_stops: [],
+    };
+  }
+
+  initialize(attributes: any, options: any) {
+    super.initialize(attributes, options);
+
+    this.createGradient();
+
+    for (const colorStop of this.get('color_stops')) {
+      this.value.addColorStop(colorStop[0], colorStop[1]);
+    }
+  }
+
+  protected createGradient() {
+    this.value = GradientModel.ctx.createLinearGradient(
+      this.get('x0'), this.get('y0'),
+      this.get('x1'), this.get('y1')
+    );
+  }
+
+  value: CanvasGradient;
+
+  static model_module = MODULE_NAME;
+  static model_module_version = MODULE_VERSION;
+
+  // Global context for creating the gradients
+  static ctx: CanvasRenderingContext2D = getContext(document.createElement('canvas'));
+}
+
+
+export
+class LinearGradientModel extends GradientModel {
+  defaults() {
+    return {...super.defaults(),
+      _model_name: LinearGradientModel.model_name,
+    };
+  }
+
+  static model_name = 'LinearGradientModel';
+}
+
+
+export
+class RadialGradientModel extends GradientModel {
+  defaults() {
+    return {...super.defaults(),
+      _model_name: RadialGradientModel.model_name,
+      r0: 0.,
+      r1: 0.,
+    };
+  }
+
+  protected createGradient() {
+    this.value = GradientModel.ctx.createRadialGradient(
+      this.get('x0'), this.get('y0'), this.get('r0'),
+      this.get('x1'), this.get('y1'), this.get('r1')
+    );
+  }
+
+  static model_name = 'RadialGradientModel';
 }
 
 
@@ -95,6 +253,13 @@ class CanvasModel extends DOMWidgetModel {
     }
   }
 
+  static ATTRS = [
+    'fillStyle', 'strokeStyle', 'globalAlpha', 'font', 'textAlign',
+    'textBaseline', 'direction', 'globalCompositeOperation',
+    'lineWidth', 'lineCap', 'lineJoin', 'miterLimit', 'lineDashOffset',
+    'shadowOffsetX', 'shadowOffsetY', 'shadowBlur', 'shadowColor',
+  ];
+
   initialize(attributes: any, options: any) {
     super.initialize(attributes, options);
 
@@ -122,7 +287,10 @@ class CanvasModel extends DOMWidgetModel {
   }
 
   private async onCommand(command: any, buffers: any) {
-    await this.processCommand(command, buffers);
+    // Retrieve the commands buffer as an object (list of commands)
+    const commands = JSON.parse(Buffer.from(getTypedArray(buffers[0], command)).toString('utf-8'));
+
+    await this.processCommand(commands, buffers.slice(1, buffers.length));
 
     this.forEachView((view: CanvasView) => {
       view.updateCanvas();
@@ -133,22 +301,28 @@ class CanvasModel extends DOMWidgetModel {
   }
 
   private async processCommand(command: any, buffers: any) {
-    if (command instanceof Array) {
+    // If it's a list of commands
+    if (command instanceof Array && command[0] instanceof Array) {
       let remainingBuffers = buffers;
 
       for (const subcommand of command) {
         let subbuffers = [];
-        if (subcommand.n_buffers) {
-          subbuffers = remainingBuffers.slice(0, subcommand.n_buffers);
-          remainingBuffers = remainingBuffers.slice(subcommand.n_buffers)
+        const nBuffers: Number = subcommand[2];
+        if (nBuffers) {
+          subbuffers = remainingBuffers.slice(0, nBuffers);
+          remainingBuffers = remainingBuffers.slice(nBuffers)
         }
         await this.processCommand(subcommand, subbuffers);
       }
       return;
     }
 
-    const args: any[] = command.args;
-    switch (command.name) {
+    const name: string = COMMANDS[command[0]];
+    const args: any[] = command[1];
+    switch (name) {
+      case 'sleep':
+        await this.sleep(args[0]);
+        break;
       case 'fillRect':
         this.fillRect(args[0], args[1], args[2], args[3]);
         break;
@@ -156,10 +330,10 @@ class CanvasModel extends DOMWidgetModel {
         this.strokeRect(args[0], args[1], args[2], args[3]);
         break;
       case 'fillRects':
-        this.drawRects(command.args, buffers, this.fillRect.bind(this));
+        this.drawRects(args, buffers, this.fillRect.bind(this));
         break;
       case 'strokeRects':
-        this.drawRects(command.args, buffers, this.strokeRect.bind(this));
+        this.drawRects(args, buffers, this.strokeRect.bind(this));
         break;
       case 'fillArc':
         this.fillArc(args[0], args[1], args[2], args[3], args[4], args[5]);
@@ -168,10 +342,10 @@ class CanvasModel extends DOMWidgetModel {
         this.strokeArc(args[0], args[1], args[2], args[3], args[4], args[5]);
         break;
       case 'fillArcs':
-        this.drawArcs(command.args, buffers, this.fillArc.bind(this));
+        this.drawArcs(args, buffers, this.fillArc.bind(this));
         break;
       case 'strokeArcs':
-        this.drawArcs(command.args, buffers, this.strokeArc.bind(this));
+        this.drawArcs(args, buffers, this.strokeArc.bind(this));
         break;
       case 'fillCircle':
         this.fillCircle(args[0], args[1], args[2]);
@@ -180,35 +354,55 @@ class CanvasModel extends DOMWidgetModel {
         this.strokeCircle(args[0], args[1], args[2]);
         break;
       case 'fillCircles':
-        this.drawCircles(command.args, buffers, this.fillCircle.bind(this));
+        this.drawCircles(args, buffers, this.fillCircle.bind(this));
         break;
       case 'strokeCircles':
-        this.drawCircles(command.args, buffers, this.strokeCircle.bind(this));
+        this.drawCircles(args, buffers, this.strokeCircle.bind(this));
         break;
       case 'strokeLine':
-        this.strokeLine(command.args, buffers);
+        this.strokeLine(args, buffers);
+        break;
+      case 'strokeLines':
+        this.strokeLines(args, buffers);
+        break;
+      case 'fillPolygon':
+        this.fillPolygon(args, buffers);
+        break;
+      case 'strokePolygon':
+        this.strokePolygon(args, buffers);
         break;
       case 'fillPath':
-        await this.fillPath(command.args, buffers);
+        await this.fillPath(args, buffers);
         break;
       case 'drawImage':
-        await this.drawImage(command.args, buffers);
+        await this.drawImage(args, buffers);
         break;
       case 'putImageData':
-        this.putImageData(command.args, buffers);
+        this.putImageData(args, buffers);
         break;
       case 'set':
-        this.setAttr(command.attr, command.value);
+        await this.setAttr(args[0], args[1]);
         break;
       case 'clear':
         this.clearCanvas();
         break;
       default:
-        this.executeCommand(command.name, command.args);
+        this.executeCommand(name, args);
         break;
     }
   }
 
+  private async sleep(time: number) {
+    this.forEachView((view: CanvasView) => {
+      view.updateCanvas();
+    });
+
+    this.trigger('new-frame');
+    this.syncImageData();
+
+    await new Promise(resolve => setTimeout(resolve, time));
+  }
+
   protected fillRect(x: number, y: number, width: number, height: number) {
     this.ctx.fillRect(x, y, width, height);
   }
@@ -310,6 +504,48 @@ class CanvasModel extends DOMWidgetModel {
     this.ctx.closePath();
   }
 
+  protected strokeLines(args: any[], buffers: any) {
+    this.ctx.beginPath();
+    const points = getArg(args[0], buffers);
+
+    // Move to the first point, then create lines between points
+    this.ctx.moveTo(points.getItem(0), points.getItem(1));
+    for (let idx = 2; idx < points.length; idx += 2) {
+      this.ctx.lineTo(points.getItem(idx), points.getItem(idx + 1));
+    }
+
+    this.ctx.stroke();
+    this.ctx.closePath();
+  }
+
+  protected fillPolygon(args: any[], buffers: any) {
+    this.ctx.beginPath();
+    const points = getArg(args[0], buffers);
+
+    // Move to the first point, then create lines between points
+    this.ctx.moveTo(points.getItem(0), points.getItem(1));
+    for (let idx = 2; idx < points.length; idx += 2) {
+      this.ctx.lineTo(points.getItem(idx), points.getItem(idx + 1));
+    }
+
+    this.ctx.closePath();
+    this.ctx.fill();
+  }
+
+  protected strokePolygon(args: any[], buffers: any) {
+    this.ctx.beginPath();
+    const points = getArg(args[0], buffers);
+
+    // Move to the first point, then create lines between points
+    this.ctx.moveTo(points.getItem(0), points.getItem(1));
+    for (let idx = 2; idx < points.length; idx += 2) {
+      this.ctx.lineTo(points.getItem(idx), points.getItem(idx + 1));
+    }
+
+    this.ctx.closePath();
+    this.ctx.stroke();
+  }
+
   protected async fillPath(args: any[], buffers: any) {
     const [serializedPath] = args;
 
@@ -329,25 +565,8 @@ class CanvasModel extends DOMWidgetModel {
     }
 
     if (image.get('_model_name') == 'ImageModel') {
-      // Create the image manually instead of creating an ImageView
-      let url: string;
-      const format = image.get('format');
-      const value = image.get('value');
-      if (format !== 'url') {
-          const blob = new Blob([value], {type: `image/${format}`});
-          url = URL.createObjectURL(blob);
-      } else {
-          url = (new TextDecoder('utf-8')).decode(value.buffer);
-      }
-
-      const img = new Image();
-      return new Promise((resolve) => {
-        img.onload = () => {
-          this._drawImage(img, x, y, width, height);
-          resolve();
-        };
-        img.src = url;
-      });
+      const img = await createImageFromWidget(image);
+      this._drawImage(img, x, y, width, height);
     }
   }
 
@@ -379,8 +598,13 @@ class CanvasModel extends DOMWidgetModel {
     this.ctx.drawImage(offscreenCanvas, dx, dy);
   }
 
-  protected setAttr(attr: string, value: any) {
-    (this.ctx as any)[attr] = value;
+  protected async setAttr(attr: number, value: any) {
+    if (typeof value === 'string' && value.startsWith('IPY')) {
+      const widgetModel: GradientModel = await unpack_models(value, this.widget_manager);
+      value = widgetModel.value;
+    }
+
+    (this.ctx as any)[CanvasModel.ATTRS[attr]] = value;
   }
 
   private clearCanvas() {
@@ -434,6 +658,8 @@ class CanvasModel extends DOMWidgetModel {
 
 export
 class RoughCanvasModel extends CanvasModel {
+  static ROUGH_ATTRS: string[] = new Array(100).concat(['roughFillStyle', 'roughness', 'bowing']);
+
   defaults() {
     return {...super.defaults(),
       _model_name: RoughCanvasModel.model_name,
@@ -466,6 +692,17 @@ class RoughCanvasModel extends CanvasModel {
     this.roughCanvas.line(args[0], args[1], args[2], args[3], this.getRoughStrokeStyle());
   }
 
+  protected strokeLines(args: any[], buffers: any) {
+    const points = getArg(args[0], buffers);
+
+    const polygon: [number, number][] = [];
+    for (let idx = 0; idx < points.length; idx += 2) {
+      polygon.push([points.getItem(idx), points.getItem(idx + 1)]);
+    }
+
+    this.roughCanvas.linearPath(polygon, this.getRoughStrokeStyle());
+  }
+
   protected async fillPath(args: any[], buffers: any) {
     const [serializedPath] = args;
 
@@ -494,21 +731,36 @@ class RoughCanvasModel extends CanvasModel {
     this.roughCanvas.arc(x, y, ellipseSize, ellipseSize, start, end, false, this.getRoughStrokeStyle());
   }
 
-  protected setAttr(attr: string, value: any) {
-    switch (attr) {
-      case 'roughFillStyle':
-        this.roughFillStyle = value;
-        break;
-      case 'roughness':
-        this.roughness = value;
-        break;
-      case 'bowing':
-        this.bowing = value;
-        break;
-      default:
-        super.setAttr(attr, value);
-        break;
+  protected fillPolygon(args: any[], buffers: any) {
+    const points = getArg(args[0], buffers);
+
+    const polygon: [number, number][] = [];
+    for (let idx = 0; idx < points.length; idx += 2) {
+      polygon.push([points.getItem(idx), points.getItem(idx + 1)]);
     }
+
+    this.roughCanvas.polygon(polygon, this.getRoughFillStyle());
+  }
+
+  protected strokePolygon(args: any[], buffers: any) {
+    const points = getArg(args[0], buffers);
+
+    const polygon: [number, number][] = [];
+    for (let idx = 0; idx < points.length; idx += 2) {
+      polygon.push([points.getItem(idx), points.getItem(idx + 1)]);
+    }
+
+    this.roughCanvas.polygon(polygon, this.getRoughStrokeStyle());
+  }
+
+  protected async setAttr(attr: number, value: any) {
+    if (RoughCanvasModel.ROUGH_ATTRS[attr]) {
+      (this as any)[RoughCanvasModel.ROUGH_ATTRS[attr]] = value;
+
+      return;
+    }
+
+    await super.setAttr(attr, value);
   }
 
   private getRoughFillStyle() {
